
# THIS FILE IS GENERATED FROM PYRAD_PROC SETUP.PY
short_version = '2.1.2'
version = '2.1.2'
full_version = '2.1.2'
git_revision = 'e5e4078aa69d2ad34cab31db49e22daca5e0be35'
compile_date_time = '2025-11-13 15:27'
username = 'runner'
release = True

if not release:
    version = full_version
